# distutils: language=c++

import numpy as np
cimport numpy as np
from cython.operator cimport dereference as deref
from libc.stdlib cimport free
from libc.stdlib cimport malloc
from libc.stdint cimport uint32_t
from libcpp.string cimport string
from libcpp.vector cimport vector
from posix.fcntl cimport open
from posix.fcntl cimport O_RDONLY
from posix.unistd cimport off_t
from spectrum_utils.spectrum import MsmsSpectrum
from spectrum_utils.spectrum import FragmentAnnotation


cdef extern from 'sys/mman.h' nogil:
    void *mmap(void *addr, size_t length, int prot, int flags, int fildes,
               off_t off)
    int munmap(void *addr, size_t length)
    enum: PROT_READ
    enum: MAP_FILE
    enum: MAP_SHARED

cdef extern from 'sys/stat.h' nogil:
    cdef struct stat:
        off_t st_size
    cdef int fstat(int fd, stat *buf)

cdef extern from 'string.h' nogil:
    void *memchr(const void *s, int c, size_t n)

cdef extern from '<string>' namespace 'std' nogil:
    int stoi(const string &stri, size_t *idx, int base)
    string to_string(int val)


cdef class SplibParser:

    cdef char *_mmap
    cdef size_t _size
    cdef size_t _pos

    def __cinit__(self, char *filename):
        fd = open(filename, O_RDONLY)
        # Get total file size.
        cdef stat statbuf
        fstat(fd, &statbuf)
        self._size = statbuf.st_size
        # Memory map the spectral library file.
        self._mmap = <char*>mmap(NULL, self._size, PROT_READ, MAP_SHARED, fd,
                                 0)
        self._pos = 0

    def __dealloc__(self):
        munmap(self._mmap, self._size)

    cdef uint32_t _read_int(self) nogil:
        self._pos += 4
        return deref(<uint32_t*>(self._mmap + self._pos - 4))

    cdef double _read_double(self) nogil:
        self._pos += 8
        return deref(<double*>(self._mmap + self._pos - 8))

    cdef string _read_line(self) nogil:
        cdef size_t offset = 0
        while self._pos + offset < self._size:
            if self._mmap[self._pos + offset] == b'\n':
                offset += 1
                break
            offset += 1
        cdef string result = string(self._mmap + self._pos, offset)
        self._pos += offset
        return result

    cdef void _skip_line(self) nogil:
        cdef size_t offset = 0
        while self._pos + offset < self._size:
            if self._mmap[self._pos + offset] == b'\n':
                offset += 1
                break
            offset += 1
        self._pos += offset

    def seek_first_spectrum(self):
        with nogil:
            # Go to the start of the file and skip non-spectrum data.
            self._pos = 8
            self._skip_line()
            for _ in range(self._read_int()):
                self._skip_line()

    def read_spectrum(self, offset: int = None):
        cdef uint32_t num_peaks
        cdef float *mz
        cdef float *intensity
        cdef vector[(string, int, int)] annotation
        cdef bint is_decoy
        cdef size_t peptide_pos, peptide_len, charge_pos, charge_len

        if offset is not None and offset >= 0:
            self._pos = offset

        with nogil:
            if self._pos >= self._size:
                raise StopIteration

            spectrum_offset = self._pos
            # Identifier.
            identifier = self._read_int()
            # Peptide sequence.
            name = self._read_line()
            peptide_pos = name.find(b'.') + 1
            peptide_len = name.find(b'.', peptide_pos) - peptide_pos
            peptide = name.substr(peptide_pos, peptide_len)
            charge_pos = name.find(b'/', peptide_pos + peptide_len) + 1
            charge_len = name.find(b' ', charge_pos)
            precursor_charge = stoi(name.substr(charge_pos, charge_len), NULL,
                                    10)
            # Precursor m/z.
            precursor_mz = self._read_double()
            # Status.
            self._skip_line()
            # Number of peaks.
            num_peaks = self._read_int()
            # Read all peaks.
            mz = <float*>malloc(num_peaks * sizeof(float))
            intensity = <float*>malloc(num_peaks * sizeof(float))
            for i in range(num_peaks):
                mz[i] = <float>self._read_double()
                intensity[i] = <float>self._read_double()
                annotation.push_back(parse_annotation(self._read_line()))
                self._skip_line()
            is_decoy = self._read_line().find(b' Remark=DECOY_') != <size_t>-1

        annotation_p = np.full((num_peaks,), None, object)
        for i in range(num_peaks):
            ion_type, ion_index, charge = annotation[i]
            if charge != -1:
                annotation_p[i] = FragmentAnnotation(
                    ion_type.decode(), ion_index, charge, mz[i])
        annotation.clear()

        spectrum = MsmsSpectrum(identifier, precursor_mz, precursor_charge,
                                np.asarray(<np.float32_t[:num_peaks]> mz),
                                np.asarray(<np.float32_t[:num_peaks]>
                                           intensity), annotation_p,
                                is_decoy=is_decoy)
        spectrum.peptide = peptide.decode()

        free(mz)
        free(intensity)

        return spectrum, spectrum_offset


cdef (string, int, int) parse_annotation(string raw) nogil:
    cdef size_t ion_index_end, ion_charge_end

    cdef char ion_type = raw.at(0)
    cdef int ion_index = -1
    cdef int charge = -1
    # Discard peaks that don't correspond to typical ion types.
    if ion_type == b'a' or ion_type == b'b' or ion_type == b'y':
        # The ion index is the subsequent numeric part.
        # `find_first_not_of` returns the position of the first character that
        # does not match or `npos` (-1).
        ion_index_end = raw.find_first_not_of(b'1234567890', 1)
        ion_index = stoi(raw.substr(1, ion_index_end), NULL, 10)
        # The ion is unmodified if the next character indicates the end of the
        # peak annotation or a specified charge.
        ion_charge_end = raw.find(b'/', ion_index_end)
        if ion_charge_end == ion_index_end:
            charge = 1
        elif raw.at(ion_index_end) == b'^':
            charge = stoi(raw.substr(ion_index_end + 1,
                                     ion_charge_end - ion_index_end - 1),
                          NULL, 10)
    return (string(1, ion_type), ion_index, charge)



import argparse
import textwrap

import configargparse


class NewlineTextHelpFormatter(argparse.HelpFormatter):

    def _fill_text(self, text, width, indent):
        return '\n'.join(
            textwrap.fill(line, width, initial_indent=indent,
                          subsequent_indent=indent,
                          replace_whitespace=False).strip()
            for line in text.splitlines(keepends=True))


class Config:
    """
    Spectral library configuration.

    Configuration settings can be specified in a config.ini file (by default in
    the working directory), or as command-line arguments.
    """

    def __init__(self) -> None:
        """
        Initialize the configuration settings and provide sensible default
        values if possible.
        """

        self._parser = configargparse.ArgParser(
            description='ANN-SoLo: Approximate nearest neighbor spectral '
                        'library searching\n'
                        '================================================'
                        '=================\n\n'
                        'Bittremieux et al. Fast open modification spectral '
                        'library searching through approximate nearest '
                        'neighbor indexing. Journal of Proteome Research 17, '
                        '3464-3474 (2018).\n\nBittremieux et al. Extremely '
                        'fast and accurate open modification spectral library '
                        'searching of high-resolution mass spectra using '
                        'feature hashing and graphics processing units. '
                        'bioRxiv (2019).\n\n'
                        'Official code website: '
                        'https://github.com/bittremieux/ANN-SoLo\n\n',
            default_config_files=['config.ini'],
            args_for_setting_config_path=['-c', '--config'],
            formatter_class=NewlineTextHelpFormatter)

        # IO
        self._parser.add_argument(
            'spectral_library_filename', help='spectral library file '
                                              '(supported formats: splib)')
        self._parser.add_argument(
            'query_filename', help='query file (supported formats: mgf)')
        self._parser.add_argument(
            'out_filename',
            help='name of the mzTab output file containing the search results')

        # PREPROCESSING
        # Spectral library resolution to round mass values.
        self._parser.add_argument(
            '--resolution', default=None, type=int,
            help='spectral library resolution; masses will be rounded to the '
                 'given number of decimals (default: no rounding)')

        # Minimum and maximum fragment peak mass values.
        self._parser.add_argument(
            '--min_mz', default=11, type=int,
            help='minimum m/z value (inclusive, default: %(default)s m/z)')
        self._parser.add_argument(
            '--max_mz', default=2010, type=int,
            help='maximum m/z value (inclusive, default: %(default)s m/z)')

        # Remove peaks around the precursor mass from fragment spectra.
        self._parser.add_argument(
            '--remove_precursor', action='store_true',
            help='remove peaks around the precursor mass '
                 '(default: no peaks are removed)')
        self._parser.add_argument(
            '--remove_precursor_tolerance', default=0, type=float,
            help='the window (in m/z) around the precursor mass to remove '
                 'peaks (default: %(default)s m/z)')

        # Minimum fragment peak intensity to filter out noise peaks.
        self._parser.add_argument(
            '--min_intensity', default=0.01, type=float,
            help='remove peaks with a lower intensity relative to the maximum '
                 'intensity (default: %(default)s)')

        # Minimum number of fragment peaks or mass range (m/z).
        self._parser.add_argument(
            '--min_peaks', default=10, type=int,
            help='discard spectra with less peaks (default: %(default)s)')
        self._parser.add_argument(
            '--min_mz_range', default=250, type=float,
            help='discard spectra with a smaller mass range '
                 '(default: %(default)s m/z)')

        # Maximum number of fragment peaks to use for each query spectrum.
        self._parser.add_argument(
            '--max_peaks_used', default=50, type=int,
            help='only use the specified most intense peaks for the query '
                 'spectra (default: %(default)s)')

        # Maximum number of fragment peaks to use for each library spectrum.
        self._parser.add_argument(
            '--max_peaks_used_library', default=50, type=int,
            help='only use the specified most intense peaks for the library '
                 'spectra (default: %(default)s)')

        # Manner in which to scale the peak intensities.
        self._parser.add_argument(
            '--scaling', default='rank', type=str,
            choices=['sqrt', 'rank'],
            help='to reduce the influence of very intense peaks, scale the'
                 ' peaks by their square root or by their rank '
                 '(default: %(default)s)')

        # MATCHING
        # Maximum SSM precursor mass tolerance.
        self._parser.add_argument(
            '--precursor_tolerance_mass', type=float, required=True,
            help='precursor mass tolerance (small window for the first level '
                 'of the cascade search)')
        self._parser.add_argument(
            '--precursor_tolerance_mode', type=str, choices=['Da', 'ppm'],
            required=True,
            help='precursor mass tolerance unit (options: %(choices)s)')
        self._parser.add_argument(
            '--precursor_tolerance_mass_open', type=float,
            help='precursor mass tolerance (wide window for the second level '
                 'of the cascade search)')
        self._parser.add_argument(
            '--precursor_tolerance_mode_open', type=str,
            choices=['Da', 'ppm'],
            help='precursor mass tolerance unit (options: %(choices)s)')

        # Fragment peak matching.
        self._parser.add_argument(
            '--fragment_mz_tolerance', type=float, required=True,
            help='fragment mass tolerance (m/z)')

        # Shifted dot product.
        self._parser.add_argument(
            '--allow_peak_shifts', action='store_true',
            help='use the shifted dot product instead of the standard dot '
                 'product')

        # Maximum FDR.
        self._parser.add_argument(
            '--fdr', default=0.01, type=float,
            help='FDR threshold to accept identifications during the cascade '
                 'search (default: %(default)s)')

        self._parser.add_argument(
            '--fdr_tolerance_mass', default=0.1, type=float,
            help='mass difference bin width for the group FDR calculation '
                 'during the second cascade level (default: %(default)s Da)')
        self._parser.add_argument(
            '--fdr_tolerance_mode', default='Da', type=str,
            choices=['Da', 'ppm'],
            help='mass difference bin unit for the group FDR calculation '
                 'during the second cascade level (default: %(default)s)')
        self._parser.add_argument(
            '--fdr_min_group_size', default=20, type=int,
            help='minimum group size for the group FDR calculation during the '
                 'second cascade level (default: %(default)s)')

        # MODE
        # Use an ANN index or the conventional brute-force mode.
        self._parser.add_argument(
            '--mode', default='ann', type=str, choices=['ann', 'bf'],
            help="search using an approximate nearest neighbors or the "
                 "traditional (brute-force) mode; 'bf': brute-force, 'ann': "
                 "approximate nearest neighbors (default: %(default)s)")

        # Bin size for the ANN index (Dalton).
        self._parser.add_argument(
            '--bin_size', default=0.04, type=float,
            help='ANN vector bin width (default: %(default)s Da)')
        # ANN vector length after hashing.
        self._parser.add_argument(
            '--hash_len', default=800, type=int,
            help='ANN vector length (default: %(default)s)')

        # Number of candidates to retrieve from the ANN index for each query.
        self._parser.add_argument(
            '--num_candidates', default=1024, type=int,
            help='number of candidates to retrieve from the ANN index for each'
                 ' query (default: %(default)s), maximum 1024 when using GPU '
                 'indexing')

        # Number of query spectra to process simultaneously.
        self._parser.add_argument(
            '--batch_size', default=16384, type=int,
            help='number of query spectra to process simultaneously '
                 '(default: %(default)s)')

        # Custom FAISS parameters.
        # Number of lists in the IVF.
        self._parser.add_argument(
            '--num_list', default=256, type=int,
            help='number of partitions in the ANN index '
                 '(default: %(default)s)')
        # Number of lists to inspect during querying.
        self._parser.add_argument(
            '--num_probe', default=128, type=int,
            help='number of partitions in the ANN index to inspect during '
                 'querying (default: %(default)s), maximum 1024 when using '
                 'GPU indexing')
        # Don't try to use the GPU.
        self._parser.add_argument(
            '--no_gpu', action='store_true',
            help="don't use the GPU for ANN searching (default: GPU is used "
                 "if available)")

        # Filled in 'parse', contains the specified settings.
        self._namespace = None

    def parse(self, args_str: str = None) -> None:
        """
        Parse the configuration settings.

        Parameters
        ----------
            args_str : str
                If None, the arguments are taken from sys.argv. Arguments that
                are not explicitly specified are taken from the configuration
                file.
        """
        self._namespace = vars(self._parser.parse_args(args_str))

    def __getattr__(self, option):
        if self._namespace is None:
            raise RuntimeError('The configuration has not been initialized')
        return self._namespace[option]

    def __getitem__(self, item):
        return self.__getattr__(item)


config = Config()



import argparse
import os
import urllib.parse as urlparse

import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
import numpy as np
from spectrum_utils import plot
from spectrum_utils.spectrum import FragmentAnnotation

from ann_solo import reader
from ann_solo import spectrum_match
from ann_solo.config import config
from ann_solo.spectrum import process_spectrum


def set_matching_peaks(library_spectrum, query_spectrum):
    peak_matches = spectrum_match.get_best_match(
        query_spectrum, [library_spectrum],
        config.fragment_mz_tolerance, config.allow_peak_shifts)[2]
    query_spectrum.annotation = np.full_like(query_spectrum.mz, None, object)
    for peak_match in peak_matches:
        library_annotation = library_spectrum.annotation[peak_match[1]]
        if library_annotation is not None:
            query_spectrum.annotation[peak_match[0]] = library_annotation
        else:
            fragment_annotation = FragmentAnnotation('z', 1, 1, 0)
            fragment_annotation.ion_type = 'unknown'
            query_spectrum.annotation[peak_match[0]] =\
                library_spectrum.annotation[peak_match[1]] =\
                fragment_annotation


def main():
    # Load the cmd arguments.
    parser = argparse.ArgumentParser(
        description='Visualize spectrum–spectrum matches from your '
                    'ANN-SoLo identification results')
    parser.add_argument(
        'mztab_filename', help='Identifications in mzTab format')
    parser.add_argument(
        'query_id', help='The identifier of the query to visualize')
    args = parser.parse_args()

    # Read the mzTab file.
    metadata = {}
    with open(args.mztab_filename) as f_mztab:
        for line in f_mztab:
            line_split = line.strip().split('\t')
            if line_split[0] == 'MTD':
                metadata[line_split[1]] = line_split[2]
            else:
                break   # Metadata lines should be on top.
    ssms = reader.read_mztab_ssms(args.mztab_filename)
    # make sure the SSM ids are strings.
    ssms.index = ssms.index.map(str)

    # Recreate the search configuration.
    settings = []
    # Search settings.
    for key in metadata:
        if 'software[1]-setting' in key:
            param = metadata[key][: metadata[key].find(' ')]
            value = metadata[key][metadata[key].rfind(' ') + 1:]
            if value != 'None':
                if value != 'False':
                    settings.append('--{}'.format(param))
                if value not in ('False', 'True'):
                    settings.append(value)
    # File names.
    settings.append('dummy_spectral_library_filename')
    settings.append('dummy_query_filename')
    settings.append('dummy_output_filename')
    config.parse(' '.join(settings))

    # Retrieve information on the requested query.
    query_id = args.query_id
    query_uri = urlparse.urlparse(urlparse.unquote(
        metadata['ms_run[1]-location']))
    query_filename = os.path.abspath(os.path.join(
        query_uri.netloc, query_uri.path))
    ssm = ssms.loc[query_id]
    library_id = ssm['accession']
    library_uri = urlparse.urlparse(urlparse.unquote(ssm['database']))
    library_filename = os.path.abspath(os.path.join(
        library_uri.netloc, library_uri.path))
    score = ssm['search_engine_score[1]']

    # Read library and query spectrum.
    with reader.SpectralLibraryReader(library_filename) as lib_reader:
        library_spectrum = lib_reader.get_spectrum(library_id, True)
    query_spectrum = None
    for spec in reader.read_mgf(query_filename):
        if spec.identifier == query_id:
            query_spectrum = process_spectrum(spec, False)
            # Make sure that the precursor charge is set for query spectra
            # with a undefined precursor charge.
            query_spectrum.precursor_charge = library_spectrum.precursor_charge
            break
    # verify that the query spectrum was found
    if query_spectrum is None:
        raise ValueError('Could not find the specified query spectrum')

    # Set the matching peaks in the query spectrum to correctly color them.
    set_matching_peaks(library_spectrum, query_spectrum)
    # Modify the colors to differentiate non-matching peaks.
    plot.colors[None] = '#757575'

    # Plot the match.
    fig, ax = plt.subplots(figsize=(20, 10))

    # Plot without annotations.
    plot.mirror(query_spectrum, library_spectrum, True, False, ax)
    # Add annotations to the library spectrum.
    max_intensity = library_spectrum.intensity.max()
    for i, annotation in enumerate(library_spectrum.annotation):
        if annotation is not None and annotation.ion_type != 'unknown':
            x = library_spectrum.mz[i]
            y = -library_spectrum.intensity[i] / max_intensity
            ax.text(x, y, str(annotation),
                    color=plot.colors[annotation.ion_type], zorder=5,
                    horizontalalignment='right', verticalalignment='center',
                    rotation=90, rotation_mode='anchor')

    ax.set_ylim(-1.1, 1.05)

    ax.text(0.5, 1.06, f'{library_spectrum.peptide}, Score: {score:.3f}',
            horizontalalignment='center', verticalalignment='bottom',
            fontsize='x-large', fontweight='bold', transform=ax.transAxes)
    ax.text(0.5, 1.02, f'File: {os.path.basename(query_filename)}, '
                       f'Scan: {query_spectrum.identifier}, '
                       f'Precursor m/z: {query_spectrum.precursor_mz:.4f}, '
                       f'Library m/z: {library_spectrum.precursor_mz:.4f}, '
                       f'Charge: {query_spectrum.precursor_charge}',
            horizontalalignment='center', verticalalignment='bottom',
            fontsize='large', transform=ax.transAxes)

    plt.savefig(f'{query_id}.png', dpi=300, bbox_inches='tight')
    plt.close()


if __name__ == '__main__':
    main()



import collections
import copy
import hashlib
import json
import logging
import multiprocessing
import os
from typing import Dict
from typing import Iterator
from typing import List

import faiss
import numexpr as ne
import numpy as np
import tqdm
from spectrum_utils.spectrum import MsmsSpectrum

from ann_solo import reader
from ann_solo import spectrum_match
from ann_solo import utils
from ann_solo.config import config
from ann_solo.spectrum import process_spectrum
from ann_solo.spectrum import spectrum_to_vector
from ann_solo.spectrum import SpectrumSpectrumMatch


class SpectralLibrary:
    """
    Spectral library search engine.

    The spectral library search engine identifies unknown query spectra by
    comparing each query spectrum against candidate spectra with a known
    peptide identity in the spectral library. The query spectrum is assigned
    the peptide sequence as its best matching library spectrum.
    """

    # Hyperparameters used to initialize the spectral library.
    _hyperparameters = ['min_mz', 'max_mz', 'bin_size', 'hash_len', 'num_list']

    # File names of the ANN indices for each charge.
    _ann_filenames = {}

    # Lock to allow only a single process to access the active ANN index.
    _ann_index_lock = multiprocessing.Lock()

    def __init__(self, filename: str) -> None:
        """
        Create a spectral library from the given spectral library file.

        New ANN indexes for every charge in the spectral library are created if
        they don't exist yet for the current ANN configuration.

        Parameters
        ----------
        filename : str
            The spectral library file name.

        Raises
        ------
        FileNotFoundError: The given spectral library file wasn't found or
            isn't supported.
        """
        try:
            self._library_reader = reader.SpectralLibraryReader(
                filename, self._get_hyperparameter_hash())
            self._library_reader.open()
        except FileNotFoundError as e:
            logging.error(e)
            raise

        self._num_probe = config.num_probe
        self._num_candidates = config.num_candidates
        self._use_gpu = not config.no_gpu and faiss.get_num_gpus()
        if self._use_gpu:
            self._res = faiss.StandardGpuResources()
            # GPU indexes can only handle maximum 1024 probes and neighbors.
            # https://github.com/facebookresearch/faiss/wiki/Faiss-on-the-GPU#limitations
            if self._num_probe > 1024:
                logging.warning('Using num_probe=1024 (maximum supported '
                                'value on the GPU), %d was supplied',
                                self._num_probe)
                self._num_probe = 1024
            if self._num_candidates > 1024:
                logging.warning('Using num_candidates=1024 (maximum supported '
                                'value on the GPU), %d was supplied',
                                self._num_candidates)
                self._num_candidates = 1024

        self._current_index = None, None

        if config.mode == 'ann':
            verify_file_existence = True
            if self._library_reader.is_recreated:
                logging.warning('ANN indexes were created using '
                                'non-compatible settings')
                verify_file_existence = False
            # Check if an ANN index exists for each charge.
            base_filename = f'{os.path.splitext(filename)[0]}_' \
                            f'{self._get_hyperparameter_hash()[:7]}'
            create_ann_charges = []
            # No need to build an ANN index for infrequent precursor charges.
            ann_charges = [charge for charge, charge_info in
                           self._library_reader.spec_info['charge'].items()
                           if len(charge_info['id']) >= config.num_list]
            for charge in sorted(ann_charges):
                self._ann_filenames[charge] = f'{base_filename}_{charge}.idxann'
                if (not verify_file_existence or
                        not os.path.isfile(self._ann_filenames[charge])):
                    create_ann_charges.append(charge)
                    logging.warning('Missing ANN index for charge %d', charge)

            # Create the missing FAISS indices.
            if create_ann_charges:
                self._create_ann_indexes(create_ann_charges)

    def _get_hyperparameter_hash(self) -> str:
        """
        Get a unique string representation of the hyperparameters used to
        initialize the spectral library.

        Returns
        -------
        str
            A hexadecimal hashed string representing the initialization
            hyperparameters.
        """
        hyperparameters_bytes = json.dumps(
            {hp: config[hp] for hp in self._hyperparameters}).encode('utf-8')
        return hashlib.sha1(hyperparameters_bytes).hexdigest()

    def _create_ann_indexes(self, charges: List[int]) -> None:
        """
        Create FAISS indexes for fast ANN candidate selection.

        Parameters
        ----------
        charges : List[int]
            Charges for which a FAISS index will be created. Sufficient library
            spectra with the corresponding precursor charge should exist.
        """
        logging.debug('Add the spectra to the spectral library ANN indexes')
        # Collect vectors for all spectra per charge.
        charge_vectors = {
            charge: np.zeros((len(self._library_reader.spec_info
                                  ['charge'][charge]['id']), config.hash_len),
                             np.float32)
            for charge in charges}
        i = {charge: 0 for charge in charge_vectors.keys()}
        for lib_spectrum, _ in tqdm.tqdm(
                self._library_reader.get_all_spectra(),
                desc='Library spectra added', leave=False, unit='spectra',
                smoothing=0.1):
            charge = lib_spectrum.precursor_charge
            if charge in charge_vectors.keys():
                spectrum_to_vector(process_spectrum(lib_spectrum, True),
                                   config.min_mz, config.max_mz,
                                   config.bin_size, config.hash_len, True,
                                   charge_vectors[charge][i[charge]])
                i[charge] += 1
        # Build an individual FAISS index per charge.
        logging.info('Build the spectral library ANN indexes')
        for charge, vectors in charge_vectors.items():
            logging.debug('Create a new ANN index for charge %d', charge)
            quantizer = faiss.IndexFlatIP(config.hash_len)
            # TODO: Use HNSW as quantizer?
            #       https://github.com/facebookresearch/faiss/blob/master/benchs/bench_hnsw.py#L136
            # quantizer = faiss.IndexHNSWFlat(config.hash_len, 32)
            # quantizer.hnsw.efSearch = 64
            # ann_index -> faiss.METRIC_L2
            # ann_index.quantizer_trains_alone = 2
            ann_index = faiss.IndexIVFFlat(quantizer, config.hash_len,
                                           config.num_list,
                                           faiss.METRIC_INNER_PRODUCT)
            ann_index.train(vectors)
            ann_index.add(vectors)
            faiss.write_index(ann_index, self._ann_filenames[charge])

        logging.debug('Finished creating the spectral library ANN indexes')

    def shutdown(self) -> None:
        """
        Release any resources to gracefully shut down.
        """
        self._library_reader.close()
        if self._current_index[1] is not None:
            self._current_index[1].reset()

    def search(self, query_filename: str) -> List[SpectrumSpectrumMatch]:
        """
        Identify all unknown spectra in the given query file.

        Parameters
        ----------
        query_filename : str
            The query file name.

        Returns
        -------
        List[SpectrumSpectrumMatch]
            A list of identified `SpectrumSpectrumMatch`es between the query
            spectra and library spectra below the given FDR threshold
            (specified in the config).
        """
        logging.info('Process file %s', query_filename)

        # Read all spectra in the query file and
        # split based on their precursor charge.
        logging.debug('Read all query spectra')
        query_spectra = collections.defaultdict(list)
        for query_spectrum in tqdm.tqdm(
                reader.read_mgf(query_filename), desc='Query spectra read',
                leave=False, unit='spectra', smoothing=0.7):
            # For queries with an unknown charge, try all possible charges.
            if query_spectrum.precursor_charge is not None:
                query_spectra_charge = [query_spectrum]
            else:
                query_spectra_charge = []
                for charge in (2, 3):
                    query_spectra_charge.append(copy.copy(query_spectrum))
                    query_spectra_charge[-1].precursor_charge = charge
            for query_spectrum_charge in query_spectra_charge:
                # Discard low-quality spectra.
                if process_spectrum(query_spectrum_charge, False).is_valid:
                    (query_spectra[query_spectrum_charge.precursor_charge]
                     .append(query_spectrum_charge))

        # Identify all query spectra.
        logging.debug('Process all query spectra')
        identifications = {}
        # Cascade level 1: standard search.
        for ssm in self._search_cascade(query_spectra, 'std'):
            identifications[ssm.identifier] = ssm
        logging.info('%d spectra identified after the standard search',
                     len(identifications))
        if (config.precursor_tolerance_mass_open is not None and
                config.precursor_tolerance_mode_open is not None):
            # Collect the remaining query spectra for the second cascade level.
            for charge, query_spectra_charge in query_spectra.items():
                query_spectra[charge] = [
                    spectrum for spectrum in query_spectra_charge
                    if spectrum.identifier not in identifications]
            # Cascade level 2: open search.
            for ssm in self._search_cascade(query_spectra, 'open'):
                identifications[ssm.identifier] = ssm
            logging.info('%d spectra identified after the open search',
                         len(identifications))

        return list(identifications.values())

    def _search_cascade(self, query_spectra: Dict[int, List[MsmsSpectrum]],
                        mode: str) -> Iterator[SpectrumSpectrumMatch]:
        """
        Perform a single level of the cascade search.

        Parameters
        ----------
        query_spectra : Dict[int, List[Spectrum]]
            A dictionary with as keys the different charges and as values lists
            of query spectra for each charge.
        mode : {'std', 'open'}
            The search mode. Either 'std' for a standard search with a small
            precursor mass window, or 'open' for an open search with a wide
            precursor mass window.


        Returns
        -------
        Iterator[SpectrumSpectrumMatch]
            An iterator of spectrum-spectrum matches that are below the FDR
            threshold (specified in the config).
        """
        if mode == 'std':
            logging.debug('Identify the query spectra using a standard search '
                          '(Δm = %s %s)',
                          config.precursor_tolerance_mass,
                          config.precursor_tolerance_mode)
        elif mode == 'open':
            logging.debug('Identify the query spectra using an open search '
                          '(Δm = %s %s)',
                          config.precursor_tolerance_mass_open,
                          config.precursor_tolerance_mode_open)

        ssms = {}
        batch_size = config.batch_size
        num_spectra = sum([len(q) for q in query_spectra.values()])
        with tqdm.tqdm(desc='Query spectra processed', total=num_spectra,
                       leave=False, unit='spectra', smoothing=0.1) as pbar:
            for charge, query_spectra_charge in query_spectra.items():
                for batch_i in range(0, len(query_spectra_charge), batch_size):
                    query_spectra_batch =\
                        query_spectra_charge[batch_i:
                                             min(batch_i + batch_size,
                                                 len(query_spectra_charge))]
                    for ssm in self._search_batch(query_spectra_batch, charge,
                                                  mode):
                        # Make sure we only retain the best identification
                        # (i.e. in case of duplicated spectra
                        # if the precursor charge was unknown).
                        if (ssm is not None and
                                (ssm.identifier not in ssms or
                                 (ssm.search_engine_score >
                                  ssms[ssm.identifier].search_engine_score))):
                            ssms[ssm.identifier] = ssm
                        pbar.update(1)
        # Store the SSMs below the FDR threshold.
        logging.debug('Filter the spectrum—spectrum matches on FDR '
                      '(threshold = %s)', config.fdr)
        if mode == 'std':
            return utils.filter_fdr(ssms.values(), config.fdr)
        elif mode == 'open':
            return utils.filter_group_fdr(ssms.values(), config.fdr,
                                          config.fdr_tolerance_mass,
                                          config.fdr_tolerance_mode,
                                          config.fdr_min_group_size)

    def _search_batch(self, query_spectra: List[MsmsSpectrum],
                      charge: int, mode: str)\
            -> Iterator[SpectrumSpectrumMatch]:
        """
        Generate spectrum-spectrum matches for a batch of query spectra with
        the same precursor charge.

        Parameters
        ----------
        query_spectra : List[Spectrum]
            The query spectra for which spectrum-spectrum matches are
            generated.
        charge : int
            The precursor charge of the query spectra.
        mode : {'std', 'open'}
            The search mode. Either 'std' for a standard search with a small
            precursor mass window, or 'open' for an open search with a wide
            precursor mass window.

        Returns
        -------
        Iterator[SpectrumSpectrumMatch]
            An iterator of spectrum-spectrum matches for every query spectrum
            that could be successfully matched to its most similar library
            spectrum.
        """
        # Find all library candidates for each query spectrum.
        for query_spectrum, library_candidates in zip(
                query_spectra, self._get_library_candidates(
                    query_spectra, charge, mode)):
            # Find the best match candidate.
            if library_candidates:
                library_match, score, _ = spectrum_match.get_best_match(
                    query_spectrum, library_candidates,
                    config.fragment_mz_tolerance,
                    config.allow_peak_shifts)
                yield SpectrumSpectrumMatch(
                    query_spectrum, library_match, score,
                    num_candidates=len(library_candidates))

    def _get_library_candidates(self, query_spectra: List[MsmsSpectrum],
                                charge: int, mode: str)\
            -> Iterator[List[MsmsSpectrum]]:
        """
        Get the library spectra to be matched against the query spectra.

        Parameters
        ----------
        query_spectra : List[Spectrum]
            The query spectra for which library candidates are retrieved.
        charge : int
            The precursor charge of the query spectra.
        mode : {'std', 'open'}
            The search mode. Either 'std' for a standard search with a small
            precursor mass window, or 'open' for an open search with a wide
            precursor mass window.

        Returns
        -------
        Iterator[List[Spectrum]]
            An iterator of lists of library candidate spectra for each query
            spectrum.

        Raises
        ------
        ValueError: Invalid search settings:
            - Unsupported search mode (either 'std' or 'open')
            - Unsupported precursor mass tolerance mode (either 'Da' or 'ppm')
        """
        if mode == 'std':
            tol_val = config.precursor_tolerance_mass
            tol_mode = config.precursor_tolerance_mode
        elif mode == 'open':
            tol_val = config.precursor_tolerance_mass_open
            tol_mode = config.precursor_tolerance_mode_open
        else:
            raise ValueError('Unknown search mode')

        # No library matches possible.
        if charge not in self._library_reader.spec_info['charge']:
            return

        library_candidates = self._library_reader.spec_info['charge'][charge]

        # First filter: precursor m/z.
        query_mzs = np.empty((len(query_spectra), 1), float)
        for i, query_spectrum in enumerate(query_spectra):
            query_mzs[i] = query_spectrum.precursor_mz
        library_mzs = library_candidates['precursor_mz'].reshape((1, -1))
        if tol_mode == 'Da':
            candidate_filters = ne.evaluate(
                'abs(query_mzs - library_mzs) * charge <= tol_val')
        elif tol_mode == 'ppm':
            candidate_filters = ne.evaluate(
                'abs(query_mzs - library_mzs) / library_mzs * 10**6'
                '<= tol_val')
        else:
            raise ValueError('Unknown precursor tolerance mode')

        # Second filter: ANN.
        if (config.mode == 'ann' and mode == 'open' and
                charge in self._ann_filenames):
            ann_index = self._get_ann_index(charge)
            query_vectors = np.zeros((len(query_spectra), config.hash_len),
                                     np.float32)
            for i, query_spectrum in enumerate(query_spectra):
                spectrum_to_vector(
                    query_spectrum, config.min_mz, config.max_mz,
                    config.bin_size, config.hash_len, True, query_vectors[i])
            mask = np.zeros_like(candidate_filters)
            for mask_i, ann_filter in zip(mask, ann_index.search(
                    query_vectors, self._num_candidates)[1]):
                mask_i[ann_filter[ann_filter != -1]] = True
            candidate_filters = np.logical_and(candidate_filters, mask)

        # Get the library candidates that pass the filter.
        for candidate_filter in candidate_filters:
            query_candidates = []
            for idx in library_candidates['id'][candidate_filter]:
                candidate = self._library_reader.get_spectrum(idx, True)
                if candidate.is_valid:
                    query_candidates.append(candidate)
            yield query_candidates

    def _get_ann_index(self, charge: int) -> faiss.IndexIVF:
        """
        Get the ANN index for the specified charge.

        This allows on-demand loading of the ANN indices and prevents having to
        keep a large amount of data for the index into memory.
        The previously used index is cached to avoid reloading the same index
        (only a single index is cached to prevent using an excessive amount of
        memory). If no index for the specified charge is cached yet, this index
        is loaded from the disk.

        To prevent loading the same index multiple times (incurring a
        significant performance quality) it is CRUCIAL that query spectrum
        processing is partitioned by charge so the previous index can be
        reused.

        Parameters
        ----------
        charge : int
            The charge for which the ANN index is retrieved.

        Returns
        -------
        faiss.IndexIVF
            The ANN index for the specified charge.
        """
        with self._ann_index_lock:
            if self._current_index[0] != charge:
                # Release memory reserved by the previous index.
                if self._current_index[1] is not None:
                    self._current_index[1].reset()
                # Load the new index.
                logging.debug('Load the ANN index for charge %d', charge)
                index = faiss.read_index(self._ann_filenames[charge])
                if self._use_gpu:
                    co = faiss.GpuClonerOptions()
                    co.useFloat16 = True
                    index = faiss.index_cpu_to_gpu(self._res, 0, index, co)
                    index.setNumProbes(self._num_probe)
                else:
                    index.nprobe = self._num_probe
                self._current_index = charge, index

            return self._current_index[1]



import functools
import math

import mmh3
import numba as nb
import numpy as np
from spectrum_utils.spectrum import MsmsSpectrum

from ann_solo.config import config


@nb.njit
def _check_spectrum_valid(spectrum_mz: np.ndarray, min_peaks: int,
                          min_mz_range: float) -> bool:
    """
    Check whether a spectrum is of high enough quality to be used for matching.

    Parameters
    ----------
    spectrum : np.ndarray
        M/z peaks of the sspectrum whose quality is checked.

    Returns
    -------
    bool
        True if the spectrum has enough peaks covering a wide enough mass
        range, False otherwise.
    """
    return (len(spectrum_mz) >= min_peaks and
            spectrum_mz[-1] - spectrum_mz[0] >= min_mz_range)


@nb.njit
def _norm_intensity(spectrum_intensity: np.ndarray) -> np.ndarray:
    """
    Normalize spectrum peak intensities.

    Parameters
    ----------
    spectrum_intensity : np.ndarray
        The spectrum peak intensities to be normalized.

    Returns
    -------
    np.ndarray
        The normalized peak intensities.
    """
    return spectrum_intensity / np.linalg.norm(spectrum_intensity)


def process_spectrum(spectrum: MsmsSpectrum, is_library: bool) -> MsmsSpectrum:
    """
    Process the peaks of the MS/MS spectrum according to the config.

    Parameters
    ----------
    spectrum : MsmsSpectrum
        The spectrum that will be processed.
    is_library : bool
        Flag specifying whether the spectrum is a query spectrum or a library
        spectrum.

    Returns
    -------
    MsmsSpectrum
        The processed spectrum. The spectrum is also changed in-place.
    """
    if spectrum.is_processed:
        return spectrum

    min_peaks = config.min_peaks
    min_mz_range = config.min_mz_range

    spectrum = spectrum.set_mz_range(config.min_mz, config.max_mz)
    if not _check_spectrum_valid(spectrum.mz, min_peaks, min_mz_range):
        spectrum.is_valid = False
        spectrum.is_processed = True
        return spectrum
    if config.resolution is not None:
        spectrum = spectrum.round(config.resolution, 'sum')
        if not _check_spectrum_valid(spectrum.mz, min_peaks, min_mz_range):
            spectrum.is_valid = False
            spectrum.is_processed = True
            return spectrum

    if config.remove_precursor:
        spectrum = spectrum.remove_precursor_peak(
            config.remove_precursor_tolerance, 'Da', 2)
        if not _check_spectrum_valid(spectrum.mz, min_peaks, min_mz_range):
            spectrum.is_valid = False
            spectrum.is_processed = True
            return spectrum

    spectrum = spectrum.filter_intensity(
        config.min_intensity, (config.max_peaks_used_library if is_library else
                               config.max_peaks_used))
    if not _check_spectrum_valid(spectrum.mz, min_peaks, min_mz_range):
        spectrum.is_valid = False
        spectrum.is_processed = True
        return spectrum

    scaling = config.scaling
    if scaling == 'sqrt':
        scaling = 'root'
    if scaling is not None:
        spectrum = spectrum.scale_intensity(
            scaling, max_rank=(config.max_peaks_used_library if is_library else
                               config.max_peaks_used))

    spectrum.intensity = _norm_intensity(spectrum.intensity)

    # Set a flag to indicate that the spectrum has been processed to avoid
    # reprocessing of library spectra for multiple queries.
    spectrum.is_valid = True
    spectrum.is_processed = True

    return spectrum


@functools.lru_cache(maxsize=None)
def get_dim(min_mz, max_mz, bin_size):
    """
    Compute the number of bins over the given mass range for the given bin
    size.

    Args:
        min_mz: The minimum mass in the mass range (inclusive).
        max_mz: The maximum mass in the mass range (inclusive).
        bin_size: The bin size (in Da).

    Returns:
        A tuple containing (i) the number of bins over the given mass range for
        the given bin size, (ii) the highest multiple of bin size lower than
        the minimum mass, (iii) the lowest multiple of the bin size greater
        than the maximum mass. These two final values are the true boundaries
        of the mass range (inclusive min, exclusive max).
    """
    min_mz, max_mz = float(min_mz), float(max_mz)
    start_dim = min_mz - min_mz % bin_size
    end_dim = max_mz + bin_size - max_mz % bin_size
    return round((end_dim - start_dim) / bin_size), start_dim, end_dim


@functools.lru_cache(maxsize=None)
def hash_idx(bin_idx: int, hash_len: int) -> int:
    """
    Hash an integer index to fall between 0 and the given maximum hash index.

    Parameters
    ----------
    bin_idx : int
        The (unbounded) index to be hashed.
    hash_len : int
        The maximum index after hashing.

    Returns
    -------
    int
        The hashed index between 0 and `hash_len`.
    """
    return mmh3.hash(str(bin_idx), 42, signed=False) % hash_len


def spectrum_to_vector(spectrum: MsmsSpectrum, min_mz: float, max_mz: float,
                       bin_size: float, hash_len: int, norm: bool = True,
                       vector: np.ndarray = None) -> np.ndarray:
    """
    Convert a `Spectrum` to a dense NumPy vector.

    Peaks are first discretized in to mass bins of width `bin_size` between
    `min_mz` and `max_mz`, after which they are hashed to random hash bins
    in the final vector.

    Parameters
    ----------
    spectrum : Spectrum
        The `Spectrum` to be converted to a vector.
    min_mz : float
        The minimum m/z to include in the vector.
    max_mz : float
        The maximum m/z to include in the vector.
    bin_size : float
        The bin size in m/z used to divide the m/z range.
    hash_len : int
        The length of the hashed vector, None if no hashing is to be done.
    norm : bool
        Normalize the vector to unit length or not.
    vector : np.ndarray, optional
        A pre-allocated vector.

    Returns
    -------
    np.ndarray
        The hashed spectrum vector with unit length.
    """
    vec_len, min_bound, _ = get_dim(min_mz, max_mz, bin_size)
    if vector is None:
        if hash_len is not None:
            vec_len = hash_len
        vector = np.zeros((vec_len,), np.float32)
        if vec_len != vector.shape[0]:
            raise ValueError('Incorrect vector dimensionality')

    for mz, intensity in zip(spectrum.mz, spectrum.intensity):
        bin_idx = math.floor((mz - min_bound) // bin_size)
        if hash_len is not None:
            bin_idx = hash_idx(bin_idx, hash_len)
        vector[bin_idx] += intensity

    if norm:
        vector /= np.linalg.norm(vector)
    return vector


class SpectrumSpectrumMatch:

    def __init__(self, query_spectrum: MsmsSpectrum,
                 library_spectrum: MsmsSpectrum = None,
                 search_engine_score: float = math.nan,
                 q: float = math.nan,
                 num_candidates: int = 0):
        self.query_spectrum = query_spectrum
        self.library_spectrum = library_spectrum
        self.search_engine_score = search_engine_score
        self.q = q
        self.num_candidates = num_candidates

    @property
    def sequence(self):
        return (self.library_spectrum.peptide
                if self.library_spectrum is not None else None)

    @property
    def identifier(self):
        return self.query_spectrum.identifier

    @property
    def index(self):
        return self.query_spectrum.index

    @property
    def accession(self):
        return (self.library_spectrum.identifier
                if self.library_spectrum is not None else None)

    @property
    def retention_time(self):
        return self.query_spectrum.retention_time

    @property
    def charge(self):
        return self.query_spectrum.precursor_charge

    @property
    def exp_mass_to_charge(self):
        return self.query_spectrum.precursor_mz

    @property
    def calc_mass_to_charge(self):
        return (self.library_spectrum.precursor_mz
                if self.library_spectrum is not None else None)

    @property
    def is_decoy(self):
        return (self.library_spectrum.is_decoy
                if self.library_spectrum is not None else None)



# distutils: language=c++
# distutils: sources=SpectrumMatch.cpp

import numpy as np
cimport numpy as np
from libcpp cimport bool as bool_t
from libcpp.utility cimport pair
from libcpp.vector cimport vector


cdef extern from 'SpectrumMatch.h' namespace 'ann_solo' nogil:
    cdef cppclass Spectrum:
        Spectrum(double, unsigned int, unsigned int,
                 np.float32_t*, np.float32_t*, np.uint8_t*) except +

    cdef cppclass SpectrumSpectrumMatch:
        SpectrumSpectrumMatch(unsigned int) nogil except +
        unsigned int getCandidateIndex() nogil
        double getScore() nogil
        vector[pair[uint, uint]]* getPeakMatches() nogil

    cdef cppclass SpectrumMatcher:
        SpectrumMatcher() nogil except +
        SpectrumSpectrumMatch* dot(
                Spectrum*, vector[Spectrum*], double, bool_t) nogil


def get_best_match(query, candidates, fragment_mz_tolerance, allow_shift):
    """
    Find the best matching candidate spectrum compared to the given query
    spectrum.

    Spectrum similarity is evaluated based on the dot product. Either a
    standard element-wise dot product (while taking the fragment mass tolerance
    into account), or an element-wise dot product with allowed peak shifts
    based on the precursor mass difference between the spectra is calculated.
    This shifted dot product can result in a more appropriate score when PTM(s)
    cause some peaks in the spectrum to be shifted to another mass value. When
    peak shifts are allowed, annotated peaks from the candidate spectra (i.e.
    known library spectra) will be taken into account, whereas no annotations
    are taken into account for the peaks from the query spectrum.

    Args:
        query: The query spectrum for which the most similar candidate spectrum
            is determined.
        candidates: All candidate spectra that are compared to the query
            spectrum.
        fragment_mz_tolerance: Mass tolerance indicating the window around the
            mass peaks used for matching two peaks.
        allow_shift: Allow peaks to be shifted according to the precursor mass
            difference or not.

    Returns:
        The candidate with the highest similarity compared to the query
        spectrum, their similarity score, and a list of tuples `(query_peak_id,
        candidate_peak_id)` of the matching peaks between the query spectrum
        and the optimal candidate spectrum.
    """
    cdef double fragment_mz_tolerance_c
    cdef bool_t allow_shift_c
    fragment_mz_tolerance_c = fragment_mz_tolerance
    allow_shift_c = allow_shift

    cdef vector[Spectrum*] candidates_vec
    cdef np.float32_t[:] mz, intensity
    cdef np.uint8_t[:] charge
    cdef unsigned int candidate_index
    cdef double score
    cdef vector[pair[uint, uint]] peak_matches

    try:
        # Convert the candidates.
        for candidate in candidates:
            if not hasattr(candidate, 'charge'):
                candidate.charge = np.zeros_like(
                    candidate.annotation, dtype=np.uint8)
                for index, annotation in enumerate(candidate.annotation):
                    if annotation is not None:
                        candidate.charge[index] = annotation.charge
            mz = candidate.mz
            intensity = candidate.intensity
            charge = candidate.charge
            candidates_vec.push_back(new Spectrum(
                candidate.precursor_mz, candidate.precursor_charge,
                len(candidate.mz), &mz[0], &intensity[0], &charge[0]))

        mz = query.mz
        intensity = query.intensity
        query.charge = np.zeros_like(query.mz, dtype=np.uint8)
        charge = query.charge
        query_spec = new Spectrum(
            query.precursor_mz, query.precursor_charge, len(query.mz),
            &mz[0], &intensity[0], &charge[0])

        with nogil:
            query_matcher = new SpectrumMatcher()
            result = query_matcher.dot(query_spec, candidates_vec,
                                       fragment_mz_tolerance_c, allow_shift_c)
            candidate_index = result.getCandidateIndex()
            score = result.getScore()
            peak_matches = result.getPeakMatches()[0]

        return (candidates[candidate_index], score,
                [peak_matches[i] for i in range(peak_matches.size())])
    finally:
        for i in range(candidates_vec.size()):
            del candidates_vec[i]
        candidates_vec.clear()
        del query_spec
        del query_matcher
        del result



import logging

from ann_solo import spectral_library
from ann_solo import writer
from ann_solo.config import config


def main():
    # Initialize logging.
    logging.basicConfig(format='{asctime} [{levelname}/{processName}] '
                               '{module}.{funcName} : {message}',
                        style='{', level=logging.DEBUG)

    # Load the configuration.
    config.parse()

    # Perform the search.
    spec_lib = spectral_library.SpectralLibrary(
        config.spectral_library_filename)
    identifications = spec_lib.search(config.query_filename)
    writer.write_mztab(identifications, config.out_filename,
                       spec_lib._library_reader)
    spec_lib.shutdown()

    logging.shutdown()


if __name__ == '__main__':
    main()



import collections
import logging
import os
import pickle
from functools import lru_cache
from typing import Iterator
from typing import List
from typing import Tuple

import joblib
import numpy as np
import pandas as pd
import tqdm
from pyteomics import mgf
from spectrum_utils.spectrum import MsmsSpectrum

from ann_solo.parsers import SplibParser
from ann_solo.spectrum import process_spectrum


class SpectralLibraryReader:
    """
    Read spectra from a SpectraST spectral library .splib file.
    """

    _supported_extensions = ['.splib']

    is_recreated = False

    def __init__(self, filename: str, config_hash: str = None) -> None:
        """
        Initialize the spectral library reader. Metadata for future easy access
        of the individual spectra is read from the corresponding configuration
        file.

        The configuration file contains minimally for each spectrum in the
        spectral library its precursor charge and precursor mass to quickly
        filter the spectra library. Furthermore, it also contains the settings
        used to construct this spectral library to make sure these match the
        runtime settings.

        Parameters
        ----------
        filename : str
            The file name of the spectral library.
        config_hash : str, optional
            The hash representing the current spectral library configuration.

        Raises
        ------
        FileNotFoundError
            The given spectral library file wasn't found.
        ValueError
            The configuration file wasn't found or its settings don't
            correspond to the runtime settings.
        """
        self._filename = filename
        self._config_hash = config_hash
        self._parser = None
        do_create = False

        # Test if the given spectral library file is in a supported format.
        verify_extension(self._supported_extensions, self._filename)

        logging.debug('Load the spectral library configuration')

        # Verify that the configuration file
        # corresponding to this spectral library is present.
        config_filename = self._get_config_filename()
        if not os.path.isfile(config_filename):
            # If not we should recreate this file
            # prior to using the spectral library.
            do_create = True
            logging.warning('Missing spectral library configuration file')
        else:
            # Load the configuration file.
            config_lib_filename, self.spec_info, load_hash =\
                joblib.load(config_filename)

            # Check that the same spectral library file format is used.
            if config_lib_filename != os.path.basename(self._filename):
                do_create = True
                logging.warning('The configuration corresponds to a different '
                                'file format of this spectral library')
            # Verify that the runtime settings match the loaded settings.
            if self._config_hash != load_hash:
                do_create = True
                logging.warning('The spectral library search engine was '
                                'created using non-compatible settings')

        # (Re)create the spectral library configuration
        # if it is missing or invalid.
        if do_create:
            self._create_config()

    def _get_config_filename(self) -> str:
        """
        Gets the configuration file name for the spectral library with the
        current configuration.

        Returns
        -------
        str
            The configuration file name (.spcfg file).
        """
        if self._config_hash is not None:
            return (f'{os.path.splitext(self._filename)[0]}_'
                    f'{self._config_hash[:7]}.spcfg')
        else:
            return f'{os.path.splitext(self._filename)[0]}.spcfg'

    def _create_config(self) -> None:
        """
        Create a new configuration file for the spectral library.

        The configuration file contains for each spectrum in the spectral
        library its offset for quick random-access reading, and its precursor
        m/z for filtering using a precursor mass window. Finally, it also
        contains the settings used to construct this spectral library to make
        sure these match the runtime settings.
        """
        logging.info('Create the spectral library configuration for file %s',
                     self._filename)

        self.is_recreated = True

        # Read all the spectra in the spectral library.
        temp_info = collections.defaultdict(
            lambda: {'id': [], 'precursor_mz': []})
        offsets = {}
        with self as lib_reader:
            for spectrum, offset in tqdm.tqdm(lib_reader.get_all_spectra(),
                                              desc='Library spectra read',
                                              unit='spectra'):
                # Store the spectrum information for easy retrieval.
                info_charge = temp_info[spectrum.precursor_charge]
                info_charge['id'].append(spectrum.identifier)
                info_charge['precursor_mz'].append(spectrum.precursor_mz)
                offsets[spectrum.identifier] = offset
        self.spec_info = {
            'charge': {
                charge: {
                    'id': np.asarray(charge_info['id'], np.uint32),
                    'precursor_mz': np.asarray(charge_info['precursor_mz'],
                                               np.float32)
                } for charge, charge_info in temp_info.items()},
            'offset': offsets}

        # Store the configuration.
        config_filename = self._get_config_filename()
        logging.debug('Save the spectral library configuration to file %s',
                      config_filename)
        joblib.dump(
            (os.path.basename(self._filename), self.spec_info,
             self._config_hash),
            config_filename, compress=9, protocol=pickle.DEFAULT_PROTOCOL)

    def open(self) -> None:
        self._parser = SplibParser(self._filename.encode())

    def close(self) -> None:
        if self._parser is not None:
            del self._parser

    def __enter__(self) -> 'SpectralLibraryReader':
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()

    @lru_cache(maxsize=None)
    def get_spectrum(self, spec_id: int, process_peaks: bool = False)\
            -> MsmsSpectrum:
        """
        Read the spectrum with the specified identifier from the spectral
        library file.

        Parameters
        ----------
        spec_id : int
            The identifier of the spectrum in the spectral library file.
        process_peaks : bool, optional
            Flag whether to process the spectrum's peaks or not
            (the default is false to not process the spectrum's peaks).

        Returns
        -------
        Spectrum
            The spectrum from the spectral library file with the specified
            identifier.
        """
        spectrum = self._parser.read_spectrum(
            self.spec_info['offset'][spec_id])[0]
        spectrum.is_processed = False
        if process_peaks:
            process_spectrum(spectrum, True)

        return spectrum

    def get_all_spectra(self) -> Iterator[Tuple[MsmsSpectrum, int]]:
        """
        Generates all spectra from the spectral library file.

        For each individual spectrum a tuple consisting of the spectrum and
        some additional information as a nested tuple (containing on the type
        of spectral library file) are returned.

        Returns
        -------
        Iterator[Tuple[Spectrum, int]]
            An iterator of all spectra along with their offset in the spectral
            library file.
        """
        self._parser.seek_first_spectrum()
        try:
            while True:
                spectrum, offset = self._parser.read_spectrum()
                spectrum.is_processed = False
                yield spectrum, offset
        except StopIteration:
            return

    def get_version(self) -> str:
        """
        Gives the spectral library version.

        Returns
        -------
        str
            A string representation of the spectral library version.
        """
        return 'null'


def verify_extension(supported_extensions: List[str], filename: str) -> None:
    """
    Check that the given file name has a supported extension.

    Parameters
    ----------
    supported_extensions : List[str]
        A list of supported file extensions.
    filename : str
        The file name to be checked.

    Raises
    ------
    FileNotFoundError
        If the file name does not have one of the supported extensions.
    """
    _, ext = os.path.splitext(os.path.basename(filename))
    if ext.lower() not in supported_extensions:
        logging.error('Unrecognized file format: %s', filename)
        raise FileNotFoundError(f'Unrecognized file format (supported file '
                                f'formats: {", ".join(supported_extensions)})')
    elif not os.path.isfile(filename):
        logging.error('File not found: %s', filename)
        raise FileNotFoundError(f'File {filename} does not exist')


def read_mgf(filename: str) -> Iterator[MsmsSpectrum]:
    """
    Read all spectra from the given mgf file.

    Parameters
    ----------
    filename: str
        The mgf file name from which to read the spectra.

    Returns
    -------
    Iterator[Spectrum]
        An iterator of spectra in the given mgf file.
    """
    # Test if the given file is an mzML file.
    verify_extension(['.mgf'], filename)

    # Get all query spectra.
    for i, mgf_spectrum in enumerate(mgf.read(filename)):
        # Create spectrum.
        identifier = mgf_spectrum['params']['title']
        precursor_mz = float(mgf_spectrum['params']['pepmass'][0])
        retention_time = float(mgf_spectrum['params']['rtinseconds'])
        if 'charge' in mgf_spectrum['params']:
            precursor_charge = int(mgf_spectrum['params']['charge'][0])
        else:
            precursor_charge = None

        spectrum = MsmsSpectrum(identifier, precursor_mz, precursor_charge,
                                mgf_spectrum['m/z array'],
                                mgf_spectrum['intensity array'],
                                retention_time=retention_time)
        spectrum.index = i
        spectrum.is_processed = False

        yield spectrum


def read_mztab_ssms(filename: str) -> pd.DataFrame:
    """
    Read SSMs from the given mzTab file.

    Parameters
    ----------
    filename: str
        The mzTab file name from which to read the SSMs.

    Returns
    -------
    pd.DataFrame
        A data frame containing the SSM information from the mzTab file.
    """
    verify_extension(['.mztab'], filename)

    # Skip the header lines.
    skiplines = 0
    with open(filename) as f_in:
        line = next(f_in)
        while line.split('\t', 1)[0] != 'PSH':
            line = next(f_in)
            skiplines += 1

    ssms = pd.read_csv(filename, sep='\t', header=skiplines,
                       index_col='PSM_ID')
    ssms.drop('PSH', 1, inplace=True)

    ssms['opt_ms_run[1]_cv_MS:1002217_decoy_peptide'] =\
        ssms['opt_ms_run[1]_cv_MS:1002217_decoy_peptide'].astype(bool)

    ssms.df_name = os.path.splitext(os.path.basename(filename))[0]

    return ssms



import logging
import os
import pathlib
import re
from typing import AnyStr
from typing import List
from typing import Pattern
from typing import Union

from ann_solo.config import config
from ann_solo.reader import SpectralLibraryReader
from ann_solo.spectrum import SpectrumSpectrumMatch
from . import __version__


def natural_sort_key(s: str, _nsre: Pattern[AnyStr] = re.compile('([0-9]+)'))\
        -> List[Union[str, int]]:
    """
    Key to be used for natural sorting of mixed alphanumeric strings.

    Source: https://stackoverflow.com/a/16090640

    Parameters
    ----------
    s : str
        The string to be converted to a sort key.
    _nsre : Pattern[AnyStr]
        Pattern to split the given string into alphanumeric substrings.

    Returns
    -------
    List[Union[str, int]]
        A list of separate int (numeric) and string (alphabetic) parts of the
        given string.
    """
    return [int(text) if text.isdigit() else text.lower()
            for text in re.split(_nsre, s)]


def write_mztab(identifications: List[SpectrumSpectrumMatch], filename: str,
                lib_reader: SpectralLibraryReader) -> None:
    """
    Write the given SSMs to an mzTab file.

    Parameters
    ----------
    identifications : List[SpectrumSpectrumMatch]
        The identifications to be exported.
    filename : str
        The file name of the mzTab output file. If it does not end with a
        '.mztab' extension this will be added.
    lib_reader : SpectralLibraryReader
        The spectral library reader used during identifications.
    """
    # Check if the filename contains the mztab extension and add if required.
    if os.path.splitext(filename)[1].lower() != '.mztab':
        filename += '.mztab'

    logging.info('Save identifications to file %s', filename)

    # Collect the necessary metadata.
    metadata = [
        ('mzTab-version', '1.0.0'),
        ('mzTab-mode', 'Summary'),
        ('mzTab-type', 'Identification'),
        ('mzTab-ID', f'ANN-SoLo_{filename}'),
        ('title', f'ANN-SoLo identification file "{filename}"'),
        ('description', f'Identification results of file '
                        f'"{os.path.split(config.query_filename)[1]}" against '
                        f'spectral library file '
                        f'"{os.path.split(config.spectral_library_filename)[1]}"'),
        ('software[1]', f'[MS, MS:1001456, ANN-SoLo, {__version__}]'),
        ('psm_search_engine_score[1]', '[MS, MS:1001143, search engine '
                                       'specific score for PSMs,]'),
        ('psm_search_engine_score[2]', '[MS, MS:1002354, PSM-level q-value,]'),
        ('ms_run[1]-format', '[MS, MS:1001062, Mascot MGF file,]'),
        ('ms_run[1]-location', pathlib.Path(
            os.path.abspath(config.query_filename)).as_uri()),
        ('ms_run[1]-id_format', '[MS, MS:1000774, multiple peak list nativeID '
                                'format,]'),
        ('fixed_mod[1]', '[MS, MS:1002453, No fixed modifications searched,]'),
        ('variable_mod[1]', '[MS, MS:1002454, No variable modifications '
                            'searched,]'),
        ('false_discovery_rate', f'[MS, MS:1002350, PSM-level global FDR, '
                                 f'{config.fdr}]'),
    ]

    # Add relevant configuration settings.
    config_keys = [
        'resolution', 'min_mz', 'max_mz', 'remove_precursor',
        'remove_precursor_tolerance', 'min_intensity', 'min_peaks',
        'min_mz_range', 'max_peaks_used', 'max_peaks_used_library', 'scaling',
        'precursor_tolerance_mass', 'precursor_tolerance_mode',
        'precursor_tolerance_mass_open', 'precursor_tolerance_mode_open',
        'fragment_mz_tolerance', 'allow_peak_shifts', 'fdr',
        'fdr_tolerance_mass', 'fdr_tolerance_mode', 'fdr_min_group_size',
        'mode']
    if config.mode == 'ann':
        config_keys.extend(['bin_size', 'hash_len', 'num_candidates',
                            'num_list', 'num_probe'])
    for i, key in enumerate(config_keys):
        metadata.append((f'software[1]-setting[{i}]',
                         f'{key} = {config[key]}'))

    database_version = lib_reader.get_version()

    with open(filename, 'w') as f_out:
        # Metadata section.
        for m in metadata:
            f_out.write('\t'.join(['MTD'] + list(m)) + '\n')

        # SSMs.
        f_out.write('\t'.join([
            'PSH', 'sequence', 'PSM_ID', 'accession', 'unique', 'database',
            'database_version', 'search_engine', 'search_engine_score[1]',
            'search_engine_score[2]', 'modifications', 'retention_time',
            'charge', 'exp_mass_to_charge', 'calc_mass_to_charge',
            'spectra_ref', 'pre', 'post', 'start', 'end',
            'opt_ms_run[1]_cv_MS:1002217_decoy_peptide',
            'opt_ms_run[1]_num_candidates']) + '\n')
        # SSMs sorted by their query identifier.
        for ssm in sorted(identifications,
                          key=lambda ssm: natural_sort_key(ssm.identifier)):
            f_out.write('\t'.join([
                'PSM',
                ssm.sequence,
                str(ssm.identifier),
                str(ssm.accession),
                'null',
                pathlib.Path(os.path.abspath(
                    config.spectral_library_filename)).as_uri(),
                database_version,
                '[MS, MS:1001456, ANN SoLo,]',
                str(ssm.search_engine_score),
                str(ssm.q),
                'null',
                str(ssm.retention_time),
                str(ssm.charge),
                str(ssm.exp_mass_to_charge),
                str(ssm.calc_mass_to_charge),
                f'ms_run[1]:index={ssm.index}',
                'null', 'null', 'null', 'null',
                f'{ssm.is_decoy:d}',
                str(ssm.num_candidates)]) + '\n')



import itertools
import operator
from typing import Iterator

import numpy as np
import pyteomics.auxiliary

from ann_solo.spectrum import SpectrumSpectrumMatch


def filter_fdr(ssms: Iterator[SpectrumSpectrumMatch], fdr: float = 0.01)\
        -> Iterator[SpectrumSpectrumMatch]:
    """
    Filter SSMs exceeding the given FDR.

    The following formula is used for FDR calculation: #D / #T.

    Parameters
    ----------
        ssms : Iterator[SpectrumSpectrumMatch]
            An iterator of SSMs to be filtered based on FDR.
        fdr : float
            The minimum FDR threshold for filtering.

    Returns
    -------
    Iterator[SpectrumSpectrumMatch]
        An iterator of the SSMs with an FDR below the given FDR threshold. Each
        SSM is assigned its q-value in the `q` attribute.
    """
    for _, _, q, ssm in pyteomics.auxiliary.qvalues(
            ssms, key=operator.attrgetter('search_engine_score'), reverse=True,
            is_decoy=operator.attrgetter('is_decoy'), remove_decoy=True,
            formula=1, correction=0, full_output=True):
        ssm.q = q
        if q <= fdr:
            yield ssm
        else:
            break


def filter_group_fdr(ssms: Iterator[SpectrumSpectrumMatch], fdr: float = 0.01,
                     tol_mass: float = None, tol_mode: str = None,
                     min_group_size: int = None)\
        -> Iterator[SpectrumSpectrumMatch]:
    """
    Filter SSMs exceeding the given FDR.

    Prior to FDR filtering SSMs are grouped based on their precursor mass
    difference. FDR filtering is applied separately to each common SSM group
    and combined to all uncommon SSM groups.

    Args:
        ssms : Iterator[SpectrumSpectrumMatch]
            An iterator of `SSMs to be filtered based on FDR.
        fdr : float
            The minimum FDR threshold for filtering.
        tol_mass : float, optional
            The mass range to group SSMs. If None no grouping is performed.
        tol_mode : str, optional
            The unit in which the mass range is specified ('Da' or 'ppm'). If
            None no grouping is performed.
        min_group_size : int, optional
            The minimum number of SSMs that should be present in a group for it
            to be considered common. If None no grouping is performed.

    Returns:
    Iterator[SpectrumSpectrumMatch]
        An iterator of the SSMs with an FDR below the given FDR threshold. Each
        SSM is assigned its q-value in the `q` variable.
    """
    ssms_remaining = np.asarray(sorted(
        ssms, key=operator.attrgetter('search_engine_score'), reverse=True))
    exp_masses = np.asarray([ssm.exp_mass_to_charge for ssm in ssms_remaining])
    mass_diffs = np.asarray([(ssm.exp_mass_to_charge - ssm.calc_mass_to_charge)
                             * ssm.charge for ssm in ssms_remaining])

    # Start with the highest ranked SSM.
    groups_common, groups_uncommon = [], []
    while ssms_remaining.size > 0:
        # Find all remaining PSMs within the mass difference window.
        if (tol_mass is None or tol_mode not in ('Da', 'ppm') or
                min_group_size is None):
            mask = np.full(len(ssms_remaining), True, dtype=bool)
        elif tol_mode == 'Da':
            mask = np.fabs(mass_diffs - mass_diffs[0]) <= tol_mass
        elif tol_mode == 'ppm':
            mask = (np.fabs(mass_diffs - mass_diffs[0]) / exp_masses * 10 ** 6
                    <= tol_mass)
        if np.count_nonzero(mask) >= min_group_size:
            groups_common.append(ssms_remaining[mask])
        else:
            groups_uncommon.extend(ssms_remaining[mask])
        # Exclude the selected SSMs from further selections.
        ssms_remaining = ssms_remaining[~mask]
        exp_masses = exp_masses[~mask]
        mass_diffs = mass_diffs[~mask]

    # Calculate the FDR combined for all uncommon mass difference groups
    # and separately for each common mass difference group.
    for ssm in itertools.chain(
            filter_fdr(groups_uncommon, fdr),
            *[filter_fdr(group, fdr) for group in groups_common]):
        yield ssm



import setuptools

import numpy as np

import ann_solo

try:
    import Cython.Distutils
except ImportError:
    use_cython = False
else:
    use_cython = True


DISTNAME = 'ann_solo'
VERSION = ann_solo.__version__
DESCRIPTION = 'Spectral library search engine optimized for fast open ' \
              'modification searching'
with open('README.md') as f_in:
    LONG_DESCRIPTION = f_in.read()
AUTHOR = 'Wout Bittremieux'
AUTHOR_EMAIL = 'wout.bittremieux@uantwerpen.be'
URL = 'https://github.com/bittremieux/ANN-SoLo'
LICENSE = 'Apache 2.0'


compile_args = ['-O3', '-march=native', '-ffast-math', '-fno-associative-math',
                '-std=c++14']
cython_directives = {'boundscheck': False, 'wraparound': False,
                     'initializedcheck': False, 'language_level': 3}
ext_spectrum_match = setuptools.Extension(
    'ann_solo.spectrum_match',
    ['ann_solo/spectrum_match.pyx', 'ann_solo/SpectrumMatch.cpp'],
    language='c++', extra_compile_args=compile_args,
    extra_link_args=compile_args, include_dirs=[np.get_include()])
ext_spectrum_match.cython_directives = cython_directives
ext_parsers = setuptools.Extension(
    'ann_solo.parsers', ['ann_solo/parsers.pyx'],
    language='c++', extra_compile_args=compile_args,
    extra_link_args=compile_args, include_dirs=[np.get_include()])
ext_parsers.cython_directives = cython_directives

cmdclass = {}
if use_cython:
    cmdclass.update({'build_ext': Cython.Distutils.build_ext})

setuptools.setup(
    name=DISTNAME,
    version=VERSION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url=URL,
    license=LICENSE,
    platforms=['any'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: Unix',
        'Programming Language :: C++',
        'Programming Language :: Cython',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Scientific/Engineering :: Bio-Informatics'],
    packages=['ann_solo'],
    entry_points={
        'console_scripts': ['ann_solo = ann_solo.ann_solo:main',
                            'ann_solo_plot = ann_solo.plot_ssm:main']},
    cmdclass=cmdclass,
    install_requires=[
        'ConfigArgParse',
        'Cython',
        'faiss',
        'joblib',
        'matplotlib',
        'mmh3',
        'numba',
        'numexpr',
        'numpy',
        'pandas',
        'pyteomics',
        'scipy',
        'spectrum_utils',
        'tqdm'],
    setup_requires=[
        'Cython',
        'numpy'],
    ext_modules=[ext_spectrum_match, ext_parsers],
)
